<?php

namespace App\Controller;

use App\Entity\Domain;
use App\Form\DomainType;
use App\Repository\DomainRepository;
use Doctrine\ORM\EntityManagerInterface;
use Symfony\Bundle\FrameworkBundle\Controller\AbstractController;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Routing\Annotation\Route;
use Symfony\Component\Security\Http\Attribute\IsGranted;
use Symfony\Component\Serializer\Exception\ExceptionInterface;
use Symfony\Component\Serializer\Normalizer\FormErrorNormalizer;
use Symfony\Component\Serializer\SerializerInterface;
use Symfony\Component\String\ByteString;

#[Route('/api')]
#[IsGranted('IS_AUTHENTICATED_FULLY')]
class DomainController extends AbstractController
{
    public function __construct(
        protected SerializerInterface $serializer,
        protected FormErrorNormalizer $errorNormalizer
    )
    {

    }
    
    #[Route('/domains', name: 'app_api_domain_list', methods: ['GET'])]
    public function getDomains(DomainRepository $domainRepository): JsonResponse
    {
        return new JsonResponse([
            'domains' => $this->serializer->normalize($domainRepository->findAll())
        ]);
    }

    /**
     * @throws ExceptionInterface
     * @throws \HttpInvalidParamException
     */
    #[Route('/domains', name: 'app_api_add_domain', methods: ['POST'])]
    public function addDomain(
        DomainRepository $domainRepository,
        Request $request,
        EntityManagerInterface $entityManager,
    ): JsonResponse
    {
        $data = json_decode($request->getContent(), true);
        if(json_last_error() !== JSON_ERROR_NONE) {
            return new JsonResponse([
                'errors' => [ 'Invalid payload' ]
            ], Response::HTTP_BAD_REQUEST);
        }

        $domain = new Domain();
        $domain->setOwnerToken(ByteString::fromRandom(64)->toString());
        $data['defaultSortPolicy'] = $data['default_sort_policy'] ?? Domain::SORT_POLICIES[array_key_first(Domain::SORT_POLICIES)];
        $form = $this->createForm(DomainType::class, $domain, [
            'csrf_protection' => false,
            'validation_groups' => 'create'
        ]);
        $form->submit($data);

        if(!$form->isValid()) {
            return new JsonResponse([
                'errors' => $this->errorNormalizer->normalize($form)
            ], Response::HTTP_BAD_REQUEST);
        }

        $entityManager->persist($domain);
        $entityManager->flush();

        return new JsonResponse([
            'domain' => $this->serializer->normalize($domain)
        ]);
    }

    /**
     * @throws ExceptionInterface
     */
    #[Route('/domains/{id}', name: 'app_api_get_domain', methods: ['GET'])]
    public function getDomain(
        Domain $domain
    )
    {
        return new JsonResponse([
            'domain' => $this->serializer->normalize($domain)
        ]);
    }

    #[Route('/domains/{id}', name: 'app_api_edit_domain', methods: ['PUT'])]
    public function editDomain(
        Request $request,
        Domain $domain,
    ): Response
    {
        $data = json_decode($request->getContent(), true);
        if(json_last_error() !== JSON_ERROR_NONE) {
            return new JsonResponse([
                'errors' => [ 'Invalid payload' ]
            ], Response::HTTP_BAD_REQUEST);
        }
        $form = $this->createForm(DomainType::class, $domain, [
            'csrf_protection' => false,
            'validation_groups' => 'update'
        ]);
        $form->submit($data, false);

        if(!$form->isValid()) {
            return new JsonResponse([
                'errors' => $this->errorNormalizer->normalize($form)
            ], Response::HTTP_BAD_REQUEST);
        }

        return new JsonResponse([
            'domain' => $this->serializer->normalize($domain)
        ]);
    }

    #[Route('/domains/{id}', name: 'app_api_delete_domain', methods: ['DELETE'])]
    public function deleteDomain(Request $request, Domain $domain, EntityManagerInterface $entityManager): Response
    {
        $entityManager->remove($domain);
        $entityManager->flush();
        return new JsonResponse([]);
    }
}
